import pprint

from cfn_sphere.file_loader import FileLoader
from cfn_sphere.aws.cfn import CloudFormation
from cfn_sphere.aws.ec2 import Ec2Api
from cfn_sphere.aws.kms import KMS
from cfn_sphere.exceptions import CfnSphereException
from cfn_sphere.stack_configuration.dependency_resolver import DependencyResolver
from cfn_sphere.util import get_logger


class ParameterResolver(object):
    """
    Resolves a given artifact identifier to the value of a stacks output.
    """

    def __init__(self, region="eu-west-1"):
        self.logger = get_logger()
        self.cfn = CloudFormation(region)
        self.ec2 = Ec2Api(region)
        self.kms = KMS(region)

    @staticmethod
    def convert_list_to_string(value):
        if not value:
            return ""

        value_string = ""
        for item in value:
            if value_string:
                value_string += ','
            value_string += str(item)
        return value_string

    def get_output_value(self, stack_outputs, stack, output_key):
        """
        Get value for a specific output key in format <stack-name>.<output>.
        :param output_key: str <stack-name>.<output>
        :return: str
        """
        self.logger.debug("Looking up output {0} from stack {1}".format(output_key, stack))

        try:
            artifact = stack_outputs[stack][output_key]
            return artifact
        except KeyError:
            raise CfnSphereException("Could not get a valid value for {0}.".format(output_key))

    @staticmethod
    def is_keep_value(value):
        return value.lower().startswith('|keeporuse|')

    @staticmethod
    def is_taupage_ami_reference(value):
        return value.lower() == '|latesttaupageami|'

    @staticmethod
    def is_kms(value):
        return value.lower().startswith('|kms|')

    @staticmethod
    def get_default_from_keep_value(value):
        return value.split('|', 2)[2]

    @staticmethod
    def is_file(value):
        return value.lower().startswith('|file|')

    def get_latest_value(self, key, value, stack_name):
        try:
            if self.cfn.stack_exists(stack_name):
                latest_stack_parameters = self.cfn.get_stack_parameters_dict(stack_name)
                latest_value = latest_stack_parameters.get(key, None)
                if latest_value:
                    self.logger.info("Will keep '{0}' as latest value for {1}".format(latest_value, key))
                    return latest_value
                else:
                    return self.get_default_from_keep_value(value)
            else:
                return self.get_default_from_keep_value(value)
        except Exception as e:
            raise CfnSphereException("Could not get latest value for {0}: {1}".format(key, e))

    def resolve_parameter_values(self, stack_name, stack_config):
        resolved_parameters = {}
        stack_outputs = self.cfn.get_stacks_outputs()

        for key, value in stack_config.parameters.items():

            if isinstance(value, list):
                self.logger.debug("List parameter found for {0}".format(key))
                for i, item in enumerate(value):
                    if DependencyResolver.is_parameter_reference(item):
                        referenced_stack, output_name = DependencyResolver.parse_stack_reference_value(item)
                        value[i] = str(self.get_output_value(stack_outputs, referenced_stack, output_name))

                value_string = self.convert_list_to_string(value)
                resolved_parameters[key] = value_string

            elif isinstance(value, str):

                if DependencyResolver.is_parameter_reference(value):
                    referenced_stack, output_name = DependencyResolver.parse_stack_reference_value(value)
                    resolved_parameters[key] = str(self.get_output_value(stack_outputs, referenced_stack, output_name))

                elif self.is_keep_value(value):
                    resolved_parameters[key] = str(self.get_latest_value(key, value, stack_name))

                elif self.is_taupage_ami_reference(value):
                    resolved_parameters[key] = str(self.ec2.get_latest_taupage_image_id())

                elif self.is_kms(value):
                    resolved_parameters[key] = str(self.kms.decrypt(value.split('|', 2)[2]))

                elif self.is_file(value):
                    url = value.split('|', 2)[2]
                    resolved_parameters[key] = FileLoader.get_file(url, stack_config.working_dir)

                else:
                    resolved_parameters[key] = value

            elif isinstance(value, bool):
                resolved_parameters[key] = str(value).lower()
            elif isinstance(value, (int, float)):
                resolved_parameters[key] = str(value)
            else:
                raise NotImplementedError("Cannot handle {0} type for key: {1}".format(type(value), key))

        return resolved_parameters

    @staticmethod
    def update_parameters_with_cli_parameters(parameters, cli_parameters, stack_name):
        """
        Overwrite parameters from stack_config with those provided by user by cli
        :param parameters: dict
        :param cli_parameters: dict
        :param stack_name: string
        :return: dict

        """
        if stack_name in cli_parameters.keys():
            for new_key, new_value in cli_parameters[stack_name].items():
                parameters[new_key] = new_value

        return parameters



import boto3
from botocore.exceptions import ClientError, BotoCoreError

from cfn_sphere.exceptions import CfnSphereBotoError
from cfn_sphere.exceptions import CfnSphereException
from cfn_sphere.util import with_boto_retry


class Ec2Api(object):
    def __init__(self, region="eu-west-1"):
        self.client = boto3.client('ec2', region_name=region)

    @with_boto_retry()
    def get_images(self, name_pattern):
        """
        Return list of AMIs matching the given name_pattern

        :param name_pattern: str: AMI name pattern
        :return: list(dict)
        :raise CfnSphereException:
        """
        filters = [
            {'Name': 'name', 'Values': [name_pattern]},
            {'Name': 'is-public', 'Values': ['false']},
            {'Name': 'state', 'Values': ['available']},
            {'Name': 'root-device-type', 'Values': ['ebs']}
        ]

        try:
            response = self.client.describe_images(ExecutableUsers=["self"], Filters=filters)
        except (BotoCoreError, ClientError) as e:
            raise CfnSphereBotoError(e)

        if not response['Images']:
            raise CfnSphereException("Could not find any private and available Taupage AMI")

        return response['Images']

    @staticmethod
    def get_latest_image_id(images_list):
        """
        Filter image with most recent CreationDate
        :param images_list: list(dict)
        :return str: image id
        """
        images = {image['CreationDate']: image['ImageId'] for image in images_list}
        creation_dates = list(images.keys())
        creation_dates.sort(reverse=True)
        return images[creation_dates[0]]

    @with_boto_retry()
    def get_latest_taupage_image_id(self):
        """
        Return the image id of the most recent private AMI matching the name pattern 'Taupage-AMI-*'

        :return: str: image id
        """
        taupage_images = self.get_images(name_pattern='Taupage-AMI-*')
        return self.get_latest_image_id(taupage_images)


if __name__ == "__main__":
    ec2 = Ec2Api()
    print(ec2.get_latest_taupage_image_id())



import boto3
from boto3.exceptions import Boto3Error
from botocore.exceptions import BotoCoreError, ClientError
from six.moves.urllib.parse import urlparse

from cfn_sphere.exceptions import CfnSphereBotoError
from cfn_sphere.util import with_boto_retry


class S3(object):
    def __init__(self):
        self.s3 = boto3.resource('s3')

    @staticmethod
    def _parse_url(url):
        url_components = urlparse(url)
        protocol = url_components.scheme
        bucket_name = url_components.netloc
        key = url_components.path.strip('/')
        return protocol, bucket_name, key

    @with_boto_retry()
    def get_contents_from_url(self, url):
        try:
            (_, bucket_name, key_name) = self._parse_url(url)
            s3_object = self.s3.Object(bucket_name, key_name)
            return s3_object.get(ResponseContentEncoding='utf-8')["Body"].read().decode('utf-8')
        except (Boto3Error, BotoCoreError, ClientError) as e:
            raise CfnSphereBotoError(e)



from botocore.exceptions import ClientError


class CfnSphereException(Exception):
    def __init__(self, message="", boto_exception=None):
        self.pretty_string = str(message)
        self.str = self.pretty_string
        self.boto_exception = boto_exception
        self.request_id = None
        try:
            self.request_id = boto_exception.request_id
            self.str += " (Request ID: {0})".format(self.request_id)
        except:
            pass

    def __str__(self):
        return self.str


class CfnStackActionFailedException(CfnSphereException):
    pass


class TemplateErrorException(CfnSphereException):
    pass


class NoConfigException(CfnSphereException):
    pass


class BadConfigException(CfnSphereException):
    pass


class CyclicDependencyException(CfnSphereException):
    pass


class InvalidDependencyGraphException(CfnSphereException):
    pass


class CfnSphereBotoError(CfnSphereException):
    def __init__(self, e):
        self.boto_exception = e
        self.pretty_string = str(e)
        self.is_throttling_exception = False

        if isinstance(e, ClientError):
            self.parse_boto_client_error(e)

    def parse_boto_client_error(self, e):
        response = e.response
        error = response["Error"]
        code = error["Code"]
        message = error["Message"]

        self.pretty_string = "{0}: {1}".format(code, message)

        if code == "Throttling":
            self.is_throttling_exception = True

    def __str__(self):
        return self.pretty_string



import base64

import boto3
from boto3.exceptions import Boto3Error
from botocore.exceptions import ClientError

from cfn_sphere.exceptions import CfnSphereBotoError


class KMS(object):
    def __init__(self, region="eu-west-1"):
        self.client = boto3.client('kms', region_name=region)

    def decrypt(self, encrypted_value):
        try:
            ciphertext_blob = base64.b64decode(encrypted_value.encode())
            response = self.client.decrypt(CiphertextBlob=ciphertext_blob)
            return response['Plaintext'].decode('utf-8')
        except Boto3Error as e:
            raise CfnSphereBotoError(e)

    def encrypt(self, key_id, cleartext_string):
        try:
            response = self.client.encrypt(KeyId=key_id, Plaintext=cleartext_string)
            return base64.b64encode(response['CiphertextBlob']).decode('utf-8')
        except (Boto3Error, ClientError) as e:
            raise CfnSphereBotoError(e)


if __name__ == "__main__":
    kms_client = KMS()
    ciphertext = kms_client.encrypt("my-key-id", "foo")
    print(kms_client.decrypt(ciphertext))



#!/usr/bin/env python

from setuptools import setup
from setuptools.command.install import install as _install

class install(_install):
    def pre_install_script(self):
        pass

    def post_install_script(self):
        pass

    def run(self):
        self.pre_install_script()

        _install.run(self)

        self.post_install_script()

if __name__ == '__main__':
    setup(
        name = 'cfn-sphere',
        version = '0.1.33',
        description = '''cfn-sphere AWS CloudFormation management cli''',
        long_description = '''cfn-sphere - A CLI tool intended to simplify AWS CloudFormation handling.''',
        author = "Marco Hoyer",
        author_email = "marco_hoyer@gmx.de",
        license = 'APACHE LICENSE, VERSION 2.0',
        url = 'https://github.com/cfn-sphere/cfn-sphere',
        scripts = ['scripts/cf'],
        packages = [
            'cfn_sphere',
            'cfn_sphere.aws',
            'cfn_sphere.stack_configuration',
            'cfn_sphere.template'
        ],
        py_modules = [],
        classifiers = [
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'Intended Audience :: System Administrators',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python',
            'Topic :: System :: Systems Administration'
        ],
        entry_points = {
            'console_scripts': ['cf=cfn_sphere.cli:main']
        },
        data_files = [],
        package_data = {},
        install_requires = [
            'boto3',
            'click',
            'gitpython',
            'networkx',
            'prettytable',
            'pyyaml',
            'six'
        ],
        dependency_links = [],
        zip_safe=True,
        cmdclass={'install': install},
    )
